
/* ########################################################################

   Copyright (c) : Alejandro Luque,1,34+954663083

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */

/*
 * bla.c: Programa principal de 'bla'
 * Alejandro Luque Estepa, 2000
 */

#include "stdlib.h"
#include "ctype.h"
#include "time.h"
#include "unistd.h"
#include "stdio.h"
#include "bla.h"
#include "verauth.h"

symrec *gl_symtbl = NULL;

/* Directorio de los ficheros include */
char *incdir = NULL;

int main(int, char**);

/* Crea un nuevo nodo terminal de expresion (STR o VAR) 
 * ntype = tipo de hoja (STR / VAR)
 * nval = valor del registro 
 */
exprec *newexpr(int ntype, char *nval)
{
  exprec *nexpr = xmalloc(sizeof(exprec));

  nexpr->type = ntype;
  nexpr->value.val = nval;
  nexpr->endings = NULL;
  nexpr->end = NULL;
  nexpr->next = NULL;
  return nexpr;
}

/* Toma dos expresiones y construye una lista (de concatenacion
 * o de bifurcacion con ellas
 * ntype = tipo de la lista (ENUM / CONCAT)
 * expr1, expr2 = elementos de la nueva lista (pueden ser listas o atomos)
 */
exprec *mklist(int ntype, exprec *expr1, exprec *expr2)
{
  exprec *nlist = xmalloc(sizeof(exprec));

  nlist->type = ntype;
  nlist->value.list = expr1;
  expr1->next = expr2;
  nlist->endings = NULL;
  nlist->end = NULL;
  nlist->next = NULL;
  return nlist;
}

/* Aade un nuevo elemento a una lista
 * olist = la lista
 * newexpr = el nuevo elemento 
 */
exprec *append(exprec *olist, exprec *newexpr)
{
  exprec *p = olist->value.list;

  while(p->next!=NULL) p = p->next;
  p->next = newexpr;
  return olist;
}

/* Encadena dos expresiones (unidas mediante el smbolo '+')
 * expr1, expr2 = las dos expresiones a unir
 */
exprec *exprcat(exprec *expr1, exprec *expr2)
{
  exprec *p;

  if(expr1->type != CONCAT) {
    if(expr2->type != CONCAT)
      return mklist(CONCAT, expr1, expr2);
    else {
      expr1->next = expr2->value.list;
      expr2->value.list = expr1;
      return expr2;
    }
  } else {
    if(expr2->type != CONCAT) 
      return append(expr1, expr2);
    else {
      p = append(expr1, expr2->value.list);
      free(expr2);
      return p;
    }
  }
}

/* Aade una lista de desinencias a la expresion dada
 * expr = la expresion dada.
 * endings = lista de desinencias.
 */
exprec *addendings(exprec *expr, exprec *endings)
{
  if(endings->type != ENUM)
    yyerror("Las desinencias deben ser una enumeracion");
  else expr->endings = endings;
  return expr;
}

/* Selecciona una desinencia para la expresion dada
 * expr = la expresion dada 
 * end = la desinencia  */
void selending(exprec *expr, char *end)
{
  expr->end = end;
}

/* Guarda un nuevo simbolo en la tabla de simbolos 
 * tbl = doble puntero a la tabla de simbolos
 * nkey = clave del simbolo
 * nvalue = valor asociado
 */
symrec *putvar(symrec *tbl, char *nkey, symrec *endtbl, exprec *nvalue)
{
  symrec *p = tbl, *lp = NULL, *nrec;
  while(p != NULL) {
    if(strcmp(nkey, p->key) >= 0)
      break;
    lp = p;
    p = p->next;
  }
  nrec = (symrec*) xmalloc(sizeof(symrec));
  nrec->key = nkey;
  nrec->value = nvalue;
  nrec->endtbl = endtbl;
  nrec->next = p;
  if(lp == NULL)
    return nrec;
  else {
    lp->next = nrec;
    return tbl;
  }
}

/* Llama a putvar para aadir un nuevo simbolo a una tabla de
 * desinencias. */
symrec *putending(symrec *endtbl, char *endname)
{
  char *s;
  static int endnum = 0;
  if(endnum > 9999) yyerror("Too many endings for that word");
  if(endtbl == NULL) endnum = 0;
  s = (char*) malloc(MAX_END_LEN);
  sprintf(s, "%d", endnum++);
  return putvar(endtbl, endname, NULL, newexpr(STR, s));
}
  
/* Vuelca una tabla de simbolos (solo para depuracion)
 */
void dumpsym(symrec *p)
{
  while(p != NULL){
    printf("%s->", p->key);
    p = p->next;
  }
  printf("NULL\n");
}

/* Recupera un simbolo de una tabla 
 * tbl = tabla de simbolos
 * nkey = clave del simbolo que hay que recuperar
 * endtbl = lugar donde almacenar un puntero a una tabla de desinencias.
 */
exprec *getvar(symrec *p, char *nkey, symrec **endtbl)
{
  int c = 0;
	
  while(p != NULL) {
    if((c = strcmp(nkey, p->key)) >= 0)
      break;
    p = p->next;
  }
  if(c==0) {
    if(endtbl != NULL) *endtbl = p->endtbl;
    return p->value;
  }
  else return NULL;
}

extern FILE *yyin;
int main(int argc, char **argv)
{
  int c;
  char *str, *fname, strstdin[] = "stdin";
	
  fname = strstdin;
	
  while((c = getopt(argc, argv, "vf:I:")) != -1)
    switch(c) {
    case 'f':
      yyin = fopen(optarg, "r");
      fname = optarg;
		    
      if(yyin == NULL) {
	fprintf(stderr, "Unable to open file %s.\n", optarg);
	exit(1);
      }
      break;
    case 'v':
      printf("BLA %s - %s\n", BLA_VERSION, BLA_AUTHOR);
      printf("Compiled %s %s.\n", __DATE__, __TIME__);
      exit(0);
      break;
    case 'I':
      incdir = optarg;
      break;
    case '?':
      if(isprint(optopt))
	fprintf(stderr, "Unknown option '-%c'.\n", optopt);
      else
	fprintf(stderr, "Unknown option character '\\x%x'.\n",
		optopt);
      break;
    default:
      abort();
    }
	
  for(c = optind; c < argc; c++) {
    str = (char*) xmalloc(MAX_ARG_LEN);
    if(snprintf(str, MAX_ARG_LEN, "$%d", c - optind + 1) < 0)
      fprintf(stderr, "Too many arguments: %s ignored.\n",
	      argv[c]);
    gl_symtbl = putvar(gl_symtbl, str, NULL, newexpr(STR, argv[c]));
  }	
  srand(time(0));
  yyinit(fname);
  yyparse();
  return 0;
}


/* Definimos una funcion xmalloc tal como se recomienda en el manual
 * de la glibc
 */
void *xmalloc(size_t size)
{
  register void *value = malloc(size);
  if(value == 0){
    fprintf(stderr, "Virtual memory exhausted");
    exit(1);
  }
  return value;
}
